socket = require('socket')
require('myrdb')

local server
local timer
local socks = {}
local deviceSocks = {}
local bufs = {}
local devices = {}
local alarms = {}
local operations = {}
local jsonPropertyPattern = '[{,]%s*"(.-)"%s*:%s*(.-)%s*[,}]'
local jsonDirectValuePattern = '[%[,]%s*(.-)%s*[,%]]'
local jsonObjectEndPattern = '}$'
local jsonObjectRestPattern = '^(.-})'
local openingObjectBracesPattern = '{'
local jsonArrayEndPattern = '%]$'
local jsonArrayRestPattern = '^(.-%])'
local openingArrayBracesPattern = '%['
local stringStartsPattern = '^"'
local objectStartsPattern = '^{'
local arrayStartsPattern = '^%['

local fragments = {
  ['X00'] = 'X00',
  ['X01'] = 'Temperatur_C',
  ['X02'] = 'Temperatur_F',
  ['X03'] = 'Druck',
  ['X04'] = 'Spannung',
  ['X05'] = 'Strom',
  ['X06'] = 'Beschleunigung',
  ['X07'] = 'Lage',
  ['X08'] = 'Lautstärke',
  ['X09'] = 'Feuchtigkeit',
  ['X10'] = 'Helligkeit',
}

local errorSeverities = {
  ['0'] = "MAJOR",
  ['1'] = "MINOR",
  ['2'] = "WARNING"
}

local errorTypes = {
  ['0'] = "emergency-stop",
  ['1'] = "logic-stop",
  ['2'] = "conveyor-stop",
  ['3'] = "disturbance",
  ['4'] = "warning",
  ['5'] = "info"
}

local function getChildDeviceID(serial)
   srDebug('looking for device with serial ' .. (serial or 'nil'))
   if devices[serial] then
      srDebug('device already known')
      return devices[serial]
   end
   http:clear()
   if http:post('300,' .. serial) <= 0 then return nil end
   local resp = http:response()
   srDebug('response ' .. (resp or 'nil'))
   if string.sub(resp, 1, 2) == "50" then
      srDebug('no device yet')
      http:clear()
      if http:post('357,' .. 'Christ (' .. serial .. '),' .. serial) <= 0 then
         return nil
      end
      resp = http:response()
      if string.sub(resp, 1, 3) == '890' then
         local ID = string.match(resp, '%d+,%d+,(%d+)')
         local msg = '302,'..ID..','..serial..'\n342,'..c8y.ID..','..ID
         if http:post(msg) < 0 then return nil end
         devices[serial] = ID
         alarms[ID] = {}
         return ID
      end
   elseif string.sub(resp, 1, 3) == "800" then
      srDebug('device found')
      return string.match(resp, "%d+,%d+,(%d+)")
   end
   return nil
end

local function assembleObjectValue(value, input, stop)
  if value:match(jsonObjectEndPattern) then
    return stop, input
  else
    local index
    while true do
      index = stop
      local rest
      _, stop, rest = input:find(jsonObjectRestPattern, index)
      if not rest then
        return stop, nil
      end
      if rest:match(openingObjectBracesPattern) then
        value = value .. rest
        stop = stop + 1
      else
        return stop, value .. rest
      end
    end
  end
end

local function assembleArrayValue(value, input, stop)
  if value:match(jsonArrayEndPattern) then
    return stop, input
  else
    local index
    while true do
      index = stop
      local rest
      _, stop, rest = input:find(jsonArrayRestPattern, index)
      if not rest then
        return stop, nil
      end
      if rest:match(openingArrayBracesPattern) then
        value = value .. rest
        stop = stop + 1
      else
        return stop, value .. rest
      end
    end
  end
end

local function toPrimitive(value)
  if value == 'true' then
    return true
  elseif value == 'false' then
    return false
  else
    return tonumber(value)
  end
end

local function parseJsonArray(input)
  local parsedArray = {}
  local index, stop = 1, 1
  while true do
    local value
    _, stop, value = input:find(jsonDirectValuePattern, index)
    if not value then
      srDebug('break')
      stop = index
      break
    end
    if value:match(objectStartsPattern) then
      srDebug('found object property')
      local assembledValue
      stop, assembledValue = assembleObjectValue(value, input, stop)
      if assembledValue then
        srDebug('assembled value: ' .. assembledValue)
        local json, _ = parseJson(assembledValue)
        table.insert(parsedArray, json)
      else
        return nil
      end
    elseif value:match(arrayStartsPattern) then
      srDebug('found array property')
      local assembledValue
      stop, assembledValue = assembleArrayValue(value, input, stop)
      if assembledValue then
        srDebug('assembled value: ' .. assembledValue)
        local json, _ = parseJsonArray(assembledValue)
        table.insert(parsedArray, json)
      else
        return nil
      end
    elseif value:match(stringStartsPattern) then
      srDebug('found string property')
      table.insert(parsedArray, value:sub(2, -2))
    else
      srDebug('found direct property')
      table.insert(parsedArray, toPrimitive(value))
    end
    index = stop
  end
  parsedArray['_isArray'] = true
  srDebug('returning parsed array')
  return parsedArray, stop
end

function parseJson(input)
  local parsed = {}
  local index, stop = 1, 1
  while true do
    local name, value
    _, stop, name, value = string.find(input, jsonPropertyPattern, index)
    if not name then
      srDebug('break')
      stop = index
      break
    end
    srDebug(string.format('pattern result %s, %s, %s', stop, name, value))
    if value:match(objectStartsPattern) then
      srDebug('found object property')
      local assembledValue
      stop, assembledValue = assembleObjectValue(value, input, stop)
      if assembledValue then
        srDebug('assembled value: ' .. assembledValue)
        parsed[name] = parseJson(assembledValue)
      else
        return nil
      end
    elseif value:match(arrayStartsPattern) then
      srDebug('found array property')
      local assembledValue
      stop, assembledValue = assembleArrayValue(value, input, stop)
      if assembledValue then
        srDebug('assembled value: ' .. assembledValue)
        parsed[name] = parseJsonArray(assembledValue)
      else
        return nil
      end
    elseif value:match(stringStartsPattern) then
      srDebug('found string property')
      parsed[name] = value:sub(2, -2)
    else
      srDebug('found direct property')
      parsed[name] = toPrimitive(value)
    end
    index = stop
  end
  srDebug('returning parsed json')
  return parsed, stop
end

local function isValidData(message)
  return message['id']
    and message['version'] == '1.0'
    and message['totalcount']
    and message['time']
    and message['operatingTime']
    and (message['operation_time']
      or message['error']
      or message['wash']
      or message['sensor']
      or message['status']
    )
end

local function convertTime(unixTime)
  return os.date('!%Y-%m-%dT%X', unixTime) .. '+0000'
end

local function stringify(input, quotes)
  if not quotes then
    quotes = '""'
  end
  local output
  if input['_isArray'] then
    output = '['
    for index, element in ipairs(input) do
      if index > 1 then
        output = output .. ','
      end
      if type(element) == 'table' then
        output = output .. stringify(element, quotes)
      elseif type(element) == 'string' then
        output = output .. quotes .. element .. quotes
      else
        output = output .. element
      end
    end
    output = output .. ']'
  else
    output = '{'
    local first = true
    for name, element in pairs(input) do
      if first then
        first = false
      else
        output = output .. ','
      end
      if type(element) == 'table' then
        output = output .. quotes .. tostring(name) .. quotes .. ':' .. stringify(element, quotes)
      elseif type(element) == 'string' then
        output = output .. quotes .. tostring(name) .. quotes .. ':' .. quotes .. element .. quotes
      else
        output = output .. quotes .. tostring(name) .. quotes .. ':' .. tostring(element)
      end
    end
    output = output .. '}'
  end
  return output
end

local function processWash(requests, id, message)
  if message['wash'] then
    srDebug('wash fragment found')
    for _, washEvent in ipairs(message['wash']) do
      table.insert(requests, string.format('360,%s,%s,"%s"', convertTime(washEvent['finish']), id, stringify(washEvent)))
    end
  end
end

local function mapFragment(fragment)
  fragment = fragments[fragment]
  if not fragment then
    fragment = fragments['X00']
  end
  return fragment
end

local function processSensor(requests, id, message)
  if message['sensor'] then
    srDebug('sensor fragment found')
    for _, sensorData in ipairs(message['sensor']) do
      local _, _, fragment, series = sensorData['id']:find('(.+):(.+)')
      fragment = mapFragment(fragment)
      table.insert(requests, string.format('359,%s,%s,%s,%s,%d,%s', convertTime(message['time']), id, fragment, series, sensorData['value'], fragment .. '_' .. series))
    end
  end
end

local function processStatus(requests, id, message)
  if message['status'] then
    srDebug('status fragment found')
    for _ ,statusElement in ipairs(message['status']) do
      table.insert(requests, string.format('361,%s,%s,%s', convertTime(statusElement['time']), id, statusElement['status']))
    end
  end
end

local function processOperationTime(requests, id, message)
  if message['operation_time'] then
    srDebug('operation_time fragment found')
    for _, operationElement in ipairs(message['operation_time']) do
      table.insert(requests, string.format('362,%s,%s,%s,%s,%s,%s', convertTime(message['time']), id, operationElement['id'], operationElement['switches'], operationElement['id'], operationElement['time']))
    end
  end
end

local function parseErrorId(errorId)
  if type(errorId) == 'string' then
    local _, _, severity, errorType, errorNumber = errorId:find('(.+):(.+):(.+)')
    return severity, errorType, errorNumber
  end
end

local function processError(requests, id, message)
  if message['error'] then
    srDebug('error fragment found')
    local cleared = false
    for _, errorData in ipairs(message['error']) do
      local severity, errorType, errorNumber = parseErrorId(errorData['errorID'])
      severity = errorSeverities[severity]
      errorType = errorTypes[errorType]
      local active = 0
      if errorData['start']  then
        srDebug('active')
        active = 1
      end
      if alarms[id][errorType] ~= active then
        if active == 1 then
          srDebug('cleared')
          cleared = true
        else
          table.insert(requests, string.format('334,%s,%s,%s,%s,Alarm %s of type %s occurred', convertTime(message['time']), id, errorType, severity, errorNumber, errorType))
        end
      end
      alarms[id][errorType] = active
    end
    if cleared then
      table.insert(requests, string.format('311,%s,ACTIVE', id))
      table.insert(requests, string.format('311,%s,ACKNOWLEDGED', id))
    end
  end
end

local function upload(id, message)
  local requests = {}
  table.insert(requests, string.format('358,%s,%s,%s,%s', convertTime(message['time']), id, message['totalcount'], message['operatingTime']))
  processWash(requests, id, message)
  processSensor(requests, id, message)
  processStatus(requests, id, message)
  processOperationTime(requests, id, message)
  processError(requests, id, message)
  c8y:send(table.concat(requests, '\n'))
end

local function isValidAck(message)
  return type(message['success'] or message['error']) == 'number'
end

local function processAck(message)
  local time = message['success'] or message['error']
  srDebug('time ' .. (time or 'nil'))
  local operation = operations[time]
  srDebug('operation ' .. tostring(operation))
  if operation and operation['id'] and operation['text'] then
    srDebug('operation found')
    if message['success'] then
      c8y:send(string.format('310,%s,%s,"%s",%s', operation['id'], 'SUCCESSFUL', operation['text'], 'success'))
    else
      c8y:send(string.format('310,%s,%s,"%s",%s', operation['id'], 'FAILED', operation['text'], 'error'))
    end
    operations[time] = nil
  else
    srWarning("no operation found for ack " .. (time or 'nil'))
  end
end

local function process(input)
  srDebug(input)
  local message, stop = parseJson(input)
  srDebug('message: ' .. (stringify(message, '"') or 'nil'))
  srDebug("stop: " .. (stop or 'nil'))
  if isValidData(message) then
    srDebug("valid data")
    local id = getChildDeviceID(message['id'])
    if not id then
       srError('Eifeler: get Device ID failed')
       return 1
    end
    srNotice("message for device " .. id)
    upload(id, message)
    return stop, message['time'], id
  elseif isValidAck(message) then
    srDebug('valid ack')
    processAck(message)
    return stop, nil
  end
  return stop, nil
end

function poll()
   local rds = socket.select(socks, nil, 0.1)
   if rds[server] then
      local client = server:accept()
      if client then
         client:settimeout(1)
         table.insert(socks, client)
         table.insert(bufs, '')
         srInfo('Christ: new client ' .. client:getpeername())
      else
         srWarning('Christ: accept: error')
      end
   end
   local rms = {}
   for i = 2, #socks do
      if rds[socks[i]] then
         local all, err, partial = socks[i]:receive('*a')
         if all then
            bufs[i] = bufs[i] .. all
            table.insert(rms, i)
         elseif err == 'timeout' then
            bufs[i] = bufs[i] .. partial
         else
            srWarning('Christ: error: ' .. err)
            table.insert(rms, i)
         end
         local idx, messageTime, deviceId = process(bufs[i])
         bufs[i] = string.sub(bufs[i], idx)
         if messageTime then
           srDebug('messageTime: ' .. (messageTime or 'nil'))
           socks[i]:send(string.format('{"success":%d}', messageTime))
           srDebug('response sent')
           deviceSocks[deviceId] = socks[i]
           srDebug('device socket saved')
         end
      end
   end
   for i = #rms, 1, -1 do
      table.remove(socks, rms[i])
      table.remove(bufs, rms[i])
   end
end


function addDevice(r)
   devices[r:value(3)] = r:value(2)
   alarms[r:value(2)] = {}
end


function clearAlarm(r)
   local type = r:value(3)
   srDebug('alarm type ' .. (type or 'nil'))
   local sourceId = r:value(4)
   srDebug('source id ' .. (sourceId or 'nil'))
   if alarms[sourceId] and alarms[sourceId][type] ~= 1 then
      local alarmId = r:value(2)
      srDebug('clearing alarm ' .. (alarmId or 'nil'))
      c8y:send('313,' .. r:value(2))
   end
end

local function buildCommandString(text, time)
  local commands = text:gsub(',', '","')
  if #commands > 0 then
    commands = '"' .. commands .. '"'
  end
  return string.format('{"action":[%s],"time":%d}', commands, time)
end

function handleCommandOperation(r)
  srDebug('shell command')
  local operationId = r:value(2)
  local text = r:value(3)
  local deviceId = r:value(4)
  if deviceId ~= c8y.ID then
    srNotice('shell operation for child device received')
    local time = os.time()
    operations[time] = {['id'] = operationId, ['text'] = text}
    local commandString = buildCommandString(text, time)
    srDebug('commandString: ' .. (commandString or 'nil'))
    local deviceSocket = deviceSocks[deviceId]
    if deviceSocket then
      srDebug('socket found')
      deviceSocket:send(commandString)
    else
      c8y:send(string.format('304,%s,%s,%s', operationId, 'No socket present for device'))
    end
  end
end

function init()
   c8y:addMsgHandler(891, 'addDevice')
   c8y:addMsgHandler(889, 'clearAlarm')
   c8y:addMsgHandler(892, 'handleCommandOperation')
   timer = c8y:addTimer(1000, 'poll')
   local port = rdbGetInt('service.cumulocity.christ.port', 3456)
   srNotice('Christ: listen on ' .. port)
   server = socket.bind('0.0.0.0', port)
   server:settimeout(5)
   table.insert(socks, server)
   table.insert(bufs, '')
   timer:start()
   return 0
end
