local socket = require('socket')
require('myrdb')

local timer
local ports = {10001, 10002, 10003, 10004, 10005}
local udp = {}
local devices = {}
local sep = ';'
local timerInterval = 1 --millisec.
local udpTimeout = 0.0001 --sec.
local series = {
   --current
   "JT1_MC",
   "JT2_MC",
   "JT3_MC",
   "JT4_MC",
   "JT5_MC",
   "JT6_MC",
   -- angle
   "JT1_JA",
   "JT2_JA",
   "JT3_JA",
   "JT4_JA",
   "JT5_JA",
   "JT6_JA",
   -- speed
   "JT1_JS",
   "JT2_JS",
   "JT3_JS",
   "JT4_JS",
   "JT5_JS",
   "JT6_JS",
   -- timestamp
   "H",
   "M",
   "S"
}

local function getChildDeviceID(serial)
   local port  = tonumber(serial:match("%-(%d+)"))
   if devices[port] then return devices[port] end
   local ID
   http:clear()
   if http:post('339,'..serial) <= 0 then return nil end
   local resp = http:response()
   if string.sub(resp, 1, 2) == "50" then
      http:clear()
      local robot = serial:match("%d$")
      local msg = '349,'..'Robot '..robot..' (port '..port..'),1440'
      if http:post(msg) <= 0 then return nil end
      resp = http:response()
      if string.sub(resp, 1, 3) == '885' then
         ID = string.match(resp, '%d+,%d+,(%d+)')
         local msg = '341,'..ID..','..serial..'\n342,'..c8y.ID..','..ID
         if http:post(msg) < 0 then return nil end
      end
   elseif string.sub(resp, 1, 3) == "800" then
      ID = string.match(resp, "%d+,%d+,(%d+)")
   end
   if ID then devices[port] = ID end
   return ID
end

local function cutErrors(str)
   local t = {}
   for v in str:gmatch("ERR;([^;]+)") do
      table.insert(t,v)
   end
   str = str:gsub("(ERR;[^;]+;?)","")
   return ((#t > 0) and t or nil), str
end

local function utcTime_ms()
   local s,dec = math.modf(socket.gettime())
   local ms = math.floor(dec * 1000 + 0.5)
   local d = os.date("!*t",s)
   return string.format('%04d-%02d-%02dT%02d:%02d:%02d.%03d+0000',
      d.year, d.month, d.day, d.hour, d.min, d.sec, ms)
end

local function getSeparatedValues(str, port)
   local t = {}
   for i,key in ipairs(series) do
      local value = str:match("^"..key..";([^;]+)")
         or str:match(";"..key..";([^;]+)")
      if value then
         table.insert(t, value)
      else
         table.insert(t, 0)
         c8y:send("334,"..utcTime_ms()..","..devices[port]
            ..",c8y_MissingSerie_"..key.."_Alarm,WARNING,Missing value of "
            ..key..". Replaced with 0.")
      end
   end
   return table.concat(t,',')
end

function poll()
   for port,udp_obj in pairs(udp) do
      local data = udp_obj:receive()
      if data and devices[port] then
         local t = utcTime_ms()
         local errors, cleared = cutErrors(data)
         if errors then
            for i,v in ipairs(errors) do
               c8y:send("334,"..t..","..devices[port]..",c8y_ErrorCode"..v
                  .."Alarm,CRITICAL,Error code "..v)
            end
         end
         c8y:send("364,"..t..","..devices[port]..","
            ..getSeparatedValues(cleared, port))
      end
   end
end

function init()
   for i,p in ipairs(ports) do
      local s = socket.udp()
      s:setsockname("*", p)
      s:settimeout(udpTimeout)
      udp[p]=s
      local serial = c8y.ID..'-'..p
      local ID = getChildDeviceID(serial)
      if ID then
         srInfo("KWSK Robot on port "..p.." has ID: "..ID)
      else
         srError("KWSK Could not retrieve ID for Robot on port "..p)
      end
   end
   timer = c8y:addTimer(timerInterval, 'poll')
   timer:start()
   return 0
end