require('vaisala/tables')

--class
WXT530 = {}

--body of the class
function WXT530:new()

   local private = {}
      --private properties
      private.deviceAddress = 0
      private.commProtocol = "P"
      private.nameOfDevice = "Vaisala WXT530 Weather Transmitter"
      private.serialNumber = nil
      private.softwareVersion = nil
      private.deviceID = nil
      private.socatCommand = "socat -U -T 0.5 - /dev/ttyAPP4,b9600,cs8,parenb=1,parodd=0,cstopb=0,raw,echo=0"
      private.socatCommandInit = "socat -U - /dev/ttyAPP4,b9600,cs8,parenb=1,parodd=0,cstopb=0,raw,echo=0"
      private.writeRetries = 9 --before retrieving deviceID
      private.queryCmdsTbl = queryCmdsTbl
      private.c8yID = nil


      --private methods
      function private:initialize()
         os.execute("sys -r rs485")

         -- needed if agent crushed and didn't terminate socat process correctly
         os.execute("killall socat; sleep 1")

         -- start socat and send SIGKILL for it to have serial communication set
         os.execute(private.socatCommandInit.." & ( sleep 1 && kill -9 $! )")

         private:write("?")

         private:retrieveDeviceID()

         private.writeRetries = 0 --after retrieving deviceID
      end

      --writes to serial and get response
      function private:write(data, test_begin, test_end)
         --default values of parameters
         if (test_begin == nil) then test_begin = private.deviceAddress end
         if (test_end == nil) then test_end = "\r\n" end

         local tr = commProtocolTbl[private.commProtocol].responseTerminator

         local cmd = "echo -ne \'"..data..tr.."\' > /dev/ttyAPP4; "
         cmd = cmd..private.socatCommand

         local i = 0 --attempts' count
         local isSuccess = false
         local response
         repeat
            i = i + 1
            srDebug("WEATHER Executing command, attempt "..i..":\n"..(cmd or ""))
            local file = io.popen(cmd)
            if (file ~= nil) then
               response = file:read("*a")
               file:close()
               if (response:match("^("..test_begin..")") ~= nil and
                  response:match("("..test_end..")$") ~= nil) then
                     isSuccess = true
                     break
               end
            else
               srError("WEATHER io.popen("..cmd..") returned nil")
            end
         until ((i - 1) >= private.writeRetries)

         if (isSuccess) then
            response = response:sub(1, -3) --remove '\r\n'
            srDebug("WEATHER Response:\n"..(response or ""))
         else
            srError("WEATHER Incorrect response:\n")
            return
         end

         return response
      end

      function private:changeCommProtocol(prot)
         local cmd = private.deviceAddress
         if (private.commProtocol ~= "S" and
            private.commProtocol ~= "R") then -- Not SDI-12
            cmd = cmd.."XU,M="..prot
         else
            cmd = cmd.."XXU,M="..prot
         end
         private:write(cmd)
         private.commProtocol = prot
      end

      function private:retrieveDeviceID()
         local current_comm_protocol = private.commProtocol
         private:changeCommProtocol("S") --turn to 'SDI-12' communication protocol
         local response = private:write("0I") --send Identification command

         if (response ~= nil) then
            private.softwareVersion, private.serialNumber =
               string.match(response, "VAISALA_WXT530(...)(........)")
         end

         if (private.softwareVersion ~= nil and
            private.serialNumber ~= nil) then
               private.deviceID = "VAISALA_WXT530".."_"..private.serialNumber
               rdbSet("service.cumulocity.wxt530.serialnumber", private.serialNumber)
               srInfo("WEATHER Device ID: "..(private.deviceID or ""))
         else
            srError("WEATHER Couldn't retrieve serial number of WXT530 device")
         end

         private:changeCommProtocol(current_comm_protocol) --turn back
      end

      function private:sendFragment(response, query_id, frag_id)
         local frag_tbl = private.queryCmdsTbl[query_id]["fragments"][frag_id]
         local toSendType = "c8y_"..frag_id
         local seriesToSend = nil

         for series_id, series_tbl in pairs(frag_tbl.series) do
            if (series_tbl.enabled == true) then
               local value, unit_abbrev =
                  response:match(series_tbl.abbrev.."=(%-?%d+%.?%d*)([%a#])")
               local unit_denote = series_tbl.units[unit_abbrev]
               if (unit_denote ~= nil and unit_denote:match("(alarm)") ~= nil) then
                  if(series_tbl.alarm ~= nil) then
                     local alarm_type = "c8y_"..(series_tbl["alarm"]["type"] or "WXTSensorAlarm")
                     c8y:send(table.concat({
                                             '338',
                                             utcTime(),
                                             private.c8yID,
                                             alarm_type,
                                             series_tbl["alarm"]["status"] or "ACTIVE",
                                             series_tbl["alarm"]["severity"] or "MAJOR",
                                             series_tbl["alarm"]["text"] or ""
                                          }, ','), 1)
                  end
               else
                  if (value ~= nil and unit_denote ~= nil) then
                     local tempSerie = table.concat({
                                             '348',
                                             private.c8yID,
                                             toSendType,
                                             series_id,
                                             value,
                                             unit_denote,
                                             toSendType
                                          }, ',')
                     if (seriesToSend == nil) then
                        seriesToSend = tempSerie
                     else
                        seriesToSend = seriesToSend.."\n"..tempSerie
                     end
                  end
               end
            end
         end

         return seriesToSend
      end

      function private:sendFragmentsFromResponse(serial_cmd)
         local query_id = serial_cmd
         local query_tbl = private.queryCmdsTbl[query_id]
         local fragmentsToSend = nil

         if (query_tbl.enabled == true) then
            local cmd = private.deviceAddress..query_id
            local response = private:write(cmd, cmd)
            if (response ~= nil) then
               for frag_id, frag_tbl in pairs(query_tbl.fragments) do
                  if (frag_tbl.enabled == true) then
                     local tempFragment =
                        private:sendFragment(response, query_id, frag_id)
                     if (tempFragment ~= nil) then
                        if (fragmentsToSend == nil) then
                           fragmentsToSend = tempFragment
                        else
                           fragmentsToSend = fragmentsToSend.."\n"..tempFragment
                        end
                     end
                  end
               end
            end
         end

         if (fragmentsToSend ~= nil) then
            c8y:send(fragmentsToSend, 0)
         end
      end

   private:initialize()

   local public = {}
      --public properties

      --public methods
      function public:getExternalID()
         if (private.deviceID ~= nil) then
            c8y:send('339,'..private.deviceID, 1)
         end
      end

      function public:postCreateDevice()
         if (private.serialNumber ~= nil and private.softwareVersion ~= nil) then
            local name = private.nameOfDevice.." (S/N "..private.serialNumber..")"
            c8y:send(table.concat({
                                    '340',
                                    name,
                                    private.serialNumber,
                                    private.softwareVersion,
                                    1440
                                 }, ','), 1)
         end
      end

      function public:postCreateExternalID()
         if (private.c8yID ~= nil and private.deviceID ~= nil) then
            c8y:send('341,'..private.c8yID..','..private.deviceID, 1)
         end
      end

      function public:postAddChildDevice()
         if (private.c8yID ~= nil) then
            c8y:send('342,'..c8y.ID..','..private.c8yID, 1)
         end
      end

      function public:getDeviceID()
         return private.deviceID
      end

      function public:getSerialNumber()
         return private.serialNumber
      end

      function public:set_c8yID(id)
         private.c8yID = id
      end

      function public:send_r1()
         private:sendFragmentsFromResponse("R1")
      end

      function public:send_r2()
         private:sendFragmentsFromResponse("R2")
      end

      function public:send_r3()
         private:sendFragmentsFromResponse("R3")
      end

      function public:send_r5()
         private:sendFragmentsFromResponse("R5")
      end

   setmetatable(public,self)
   self.__index = self
   return public
end
